@@ -3,27 +3,39 @@ from __future__ import generators
 True, False = 1==1, 0==1
 
 import threading
+from util.threadinglocal import local as threading_local
 import re
 import warnings
 import atexit
 import os
 import new
+import types
+import urllib
+import weakref
+import inspect
 import sqlbuilder
 from cache import CacheSet
 import col
+popKey = col.popKey
+import main
 from joins import sorter
 from converters import sqlrepr
-import urllib
+import classregistry
 
 warnings.filterwarnings("ignore", "DB-API extension cursor.lastrowid used")
 
 _connections = {}
 
+def _closeConnection(ref):
+    conn = ref()
+    if conn is not None:
+        conn.close()
+
 class DBConnection:
 
     def __init__(self, name=None, debug=False, debugOutput=False,
                  cache=True, style=None, autoCommit=True,
-                 debugThreading=False):
+                 debugThreading=False, registry=None):
         self.name = name
         self.debug = debug
         self.debugOutput = debugOutput
@@ -34,22 +46,29 @@ class DBConnection:
         self._connectionNumbers = {}
         self._connectionCount = 1
         self.autoCommit = autoCommit
+        self.registry = registry or None
+        classregistry.registry(self.registry).addCallback(
+            self.soClassAdded)
         registerConnectionInstance(self)
+        atexit.register(_closeConnection, weakref.ref(self))
 
     def uri(self):
-        auth = self.user or ''
+        auth = getattr(self, 'user', None) or ''
         if auth:
             if self.password:
-                auth = auth + '@' + self.password
-            auth = auth + ':'
+                auth = auth + ':' + self.password
+            auth = auth + '@'
         else:
-            assert not password, 'URIs cannot express passwords without usernames'
+            assert not getattr(self, 'password', None), (
+                'URIs cannot express passwords without usernames')
         uri = '%s://%s' % (self.dbName, auth)
         if self.host:
-            uri += self.host + '/'
-        if path.startswith('/'):
-            path = path[1:]
-        return uri + path
+            uri += self.host
+        uri += '/'
+        db = self.db
+        if db.startswith('/'):
+            db = path[1:]
+        return uri + db
 
     def isSupported(cls):
         raise NotImplemented
@@ -76,14 +95,29 @@ class DBConnection:
             else:
                 host, rest = rest.split('/', 1)
         if host and host.find('@') != -1:
-            user, host = host.split('@', 1)
+            user = host[:host.rfind('@')] # Python 2.3 doesn't have .rsplit()
+            host = host[host.rfind('@')+1:] # !!!
             if user.find(':') != -1:
                 user, password = user.split(':', 1)
             else:
                 password = None
         else:
             user = password = None
+        if host and host.find(':') != -1:
+            _host, port = host.split(':')
+            try:
+                port = int(port)
+            except ValueError:
+                raise ValueError, "port must be integer, got '%s' instead" % port
+            if not (1 <= port <= 65535):
+                raise ValueError, "port must be integer in the range 1-65535, got '%d' instead" % port
+            host = _host
+        else:
+            port = None
         path = '/' + rest
+        if os.name == 'nt':
+            if (len(rest) > 1) and (rest[1] == '|'):
+                path = "%s:%s" % (rest[0], rest[2:])
         args = {}
         if path.find('?') != -1:
             path, arglist = path.split('?', 1)
@@ -92,9 +126,78 @@ class DBConnection:
                 argname, argvalue = single.split('=', 1)
                 argvalue = urllib.unquote(argvalue)
                 args[argname] = argvalue
-        return user, password, host, path, args
+        return user, password, host, port, path, args
     _parseURI = staticmethod(_parseURI)
 
+    def soClassAdded(self, soClass):
+        """
+        This is called for each new class; we use this opportunity
+        to create an instance method that is bound to the class
+        and this connection.
+        """
+        name = soClass.__name__
+        assert not hasattr(self, name), (
+            "Connection %r already has an attribute with the name "
+            "%r (and you just created the conflicting class %r)"
+            % (self, name, soClass))
+        setattr(self, name, ConnWrapper(soClass, self))
+
+class ConnWrapper(object):
+
+    """
+    This represents a SQLObject class that is bound to a specific
+    connection (instances have a connection instance variable, but
+    classes are global, so this is binds the connection variable
+    lazily when a class method is accessed)
+    """
+    # @@: methods that take connection arguments should be explicitly
+    # marked up instead of the implicit use of a connection argument
+    # and inspect.getargspec()
+
+    def __init__(self, soClass, connection):
+        self._soClass = soClass
+        self._connection = connection
+
+    def __call__(self, *args, **kw):
+        kw['connection'] = self._connection
+        return self._soClass(*args, **kw)
+
+    def __getattr__(self, attr):
+        meth = getattr(self._soClass, attr)
+        if not isinstance(meth, types.MethodType):
+            # We don't need to wrap non-methods
+            return meth
+        try:
+            takes_conn = meth.takes_connection
+        except AttributeError:
+            args, varargs, varkw, defaults = inspect.getargspec(meth)
+            assert not varkw and not varargs, (
+                "I cannot tell whether I must wrap this method, "
+                "because it takes **kw: %r"
+                % meth)
+            takes_conn = 'connection' in args
+            meth.im_func.takes_connection = takes_conn
+        if not takes_conn:
+            return meth
+        return ConnMethodWrapper(meth, self._connection)
+
+class ConnMethodWrapper(object):
+
+    def __init__(self, method, connection):
+        self._method = method
+        self._connection = connection
+
+    def __getattr__(self, attr):
+        return getattr(self._method, attr)
+
+    def __call__(self, *args, **kw):
+        kw['connection'] = self._connection
+        return self._method(*args, **kw)
+
+    def __repr__(self):
+        return '<Wrapped %r with connection %r>' % (
+            self._method, self._connection)
+
 class DBAPI(DBConnection):
 
     """
@@ -110,6 +213,7 @@ class DBAPI(DBConnection):
         self._pool = []
         self._poolLock = threading.Lock()
         DBConnection.__init__(self, **kw)
+        self._binaryType = type(self.module.Binary(''))
 
     def _runWithConnection(self, meth, *args):
         conn = self.getConnection()
@@ -123,17 +227,17 @@ class DBAPI(DBConnection):
         self._poolLock.acquire()
         try:
             if not self._pool:
-                newConn = self.makeConnection()
-                self._pool.append(newConn)
-                self._connectionNumbers[id(newConn)] = self._connectionCount
+                conn = self.makeConnection()
+                self._connectionNumbers[id(conn)] = self._connectionCount
                 self._connectionCount += 1
-            val = self._pool.pop()
+            else:
+                conn = self._pool.pop()
             if self.debug:
                 s = 'ACQUIRE'
                 if self._pool is not None:
                     s += ' pool=[%s]' % ', '.join([str(self._connectionNumbers[id(v)]) for v in self._pool])
-                self.printDebug(val, s, 'Pool')
-            return val
+                self.printDebug(conn, s, 'Pool')
+            return conn
         finally:
             self._poolLock.release()
 
@@ -157,7 +261,8 @@ class DBAPI(DBConnection):
             elif self.autoCommit:
                 if self.debug:
                     self.printDebug(conn, 'auto', 'COMMIT')
-                conn.commit()
+                if not getattr(conn, 'autocommit', False):
+                    conn.commit()
             else:
                 if self.debug:
                     self.printDebug(conn, 'auto', 'ROLLBACK')
@@ -167,9 +272,13 @@ class DBAPI(DBConnection):
                 # @@: We can get duplicate releasing of connections with
                 # the __del__ in Iteration (unfortunately, not sure why
                 # it happens)
-                self._pool.append(conn)
+                self._pool.insert(0, conn)
+        else:
+            conn.close()
 
     def printDebug(self, conn, s, name, type='query'):
+        if name == 'Pool' and self.debug != 'Pool':
+            return
         if type == 'query':
             sep = ': '
         else:
@@ -185,6 +294,8 @@ class DBAPI(DBConnection):
         print '%(n)2i%(threadName)s/%(name)s%(spaces)s%(sep)s %(s)s' % locals()
 
     def _executeRetry(self, conn, cursor, query):
+        if self.debug:
+            self.printDebug(conn, query, 'QueryR')
         return cursor.execute(query)
 
     def _query(self, conn, s):
@@ -208,6 +319,23 @@ class DBAPI(DBConnection):
     def queryAll(self, s):
         return self._runWithConnection(self._queryAll, s)
 
+    def _queryAllDescription(self, conn, s):
+        """
+        Like queryAll, but returns (description, rows), where the
+        description is cursor.description (which gives row types)
+        """
+        if self.debug:
+            self.printDebug(conn, s, 'QueryAllDesc')
+        c = conn.cursor()
+        self._executeRetry(conn, c, s)
+        value = c.fetchall()
+        if self.debugOutput:
+            self.printDebug(conn, value, 'QueryAll', 'result')
+        return c.description, value
+
+    def queryAllDescription(self, s):
+        return self._runWithConnection(self._queryAllDescription, s)
+
     def _queryOne(self, conn, s):
         if self.debug:
             self.printDebug(conn, s, 'QueryOne')
@@ -233,37 +361,101 @@ class DBAPI(DBConnection):
         return self._runWithConnection(self._queryInsertID, soInstance, id, names, values)
 
     def iterSelect(self, select):
-        return Iteration(self, self.getConnection(),
+        return select.IterationClass(self, self.getConnection(),
                          select, keepConnection=False)
 
-    def accumulateSelect(self, select, expression):
-        """ Apply an accumulate function (like SUM, COUNT, ..)
+    def accumulateSelect(self, select, *expressions):
+        """ Apply an accumulate function(s) (SUM, COUNT, MIN, AVG, MAX, etc...)
             to the select object.
-            Return the value resulting from the SQL accumulate function
-            as an integer.
         """
-        q = "SELECT %s" % expression
-        q += " FROM %s WHERE" % ", ".join(select.tables)
+        ops = select.ops
+        join = ops.get('join')
+        if join:
+            tables = self._fixTablesForJoins(select)
+        else:
+            tables = select.tables
+        q = "SELECT %s" % ", ".join([str(expression) for expression in expressions])
+        q += " FROM %s" % ", ".join(tables)
+        if join:
+            q += self._addJoins(select, tables)
+        q += " WHERE"
         q = self._addWhereClause(select, q, limit=0, order=0)
-        val = int(self.queryOne(q)[0])
+        val = self.queryOne(q)
+        if len(expressions) == 1:
+            val = val[0]
         return val
 
     def queryForSelect(self, select):
         ops = select.ops
+        join = ops.get('join')
         cls = select.sourceClass
-        if ops.get('lazyColumns', 0):
-            q = "SELECT %s.%s FROM %s WHERE " % \
-                (cls._table, cls._idName,
-                 ", ".join(select.tables))
+        if join:
+            tables = self._fixTablesForJoins(select)
         else:
-            q = "SELECT %s.%s, %s FROM %s WHERE " % \
-                (cls._table, cls._idName,
-                 ", ".join(["%s.%s" % (cls._table, col.dbName)
-                            for col in cls._SO_columns]),
-                 ", ".join(select.tables))
+            tables = select.tables
+        if ops.get('distinct', False):
+            q = 'SELECT DISTINCT '
+        else:
+            q = 'SELECT '
+        if ops.get('lazyColumns', 0):
+            q += "%s.%s FROM %s" % \
+                 (cls.sqlmeta.table, cls.sqlmeta.idName,
+                  ", ".join(tables))
+        else:
+            columns = ", ".join(["%s.%s" % (cls.sqlmeta.table, col.dbName)
+                                 for col in cls.sqlmeta.columnList])
+            if columns:
+                q += "%s.%s, %s FROM %s" % \
+                     (cls.sqlmeta.table, cls.sqlmeta.idName, columns,
+                      ", ".join(tables))
+            else:
+                q += "%s.%s FROM %s" % \
+                     (cls.sqlmeta.table, cls.sqlmeta.idName,
+                      ", ".join(tables))
 
+        if join:
+            q += self._addJoins(select, tables)
+        q += " WHERE"
         return self._addWhereClause(select, q)
 
+    def _fixTablesForJoins(self, select):
+        ops = select.ops
+        join = ops.get('join')
+        tables = select.tables
+        if type(join) is str:
+            return tables
+        else:
+            tables = tables[:] # maka a copy for modification
+            if isinstance(join, sqlbuilder.SQLJoin):
+                if join.table1 in tables: tables.remove(join.table1)
+                if join.table2 in tables: tables.remove(join.table2)
+            else:
+                for j in join:
+                    if j.table1 in tables: tables.remove(j.table1)
+                    if j.table2 in tables: tables.remove(j.table2)
+            return tables
+
+    def _addJoins(self, select, tables):
+        ops = select.ops
+        join = ops.get('join')
+        if type(join) is str:
+            join_str = ' ' + join
+        elif isinstance(join, sqlbuilder.SQLJoin):
+            if tables and join.table1:
+                join_str = ", "
+            else:
+                join_str = ' '
+            join_str += self.sqlrepr(join)
+        else:
+            join_str = ""
+            for j in join:
+                if tables and j.table1:
+                    sep = ", "
+                else:
+                    sep = ' '
+                join_str += sep + self.sqlrepr(j)
+        return join_str
+
     def _addWhereClause(self, select, startSelect, limit=1, order=1):
 
         q = select.clause
@@ -287,8 +479,8 @@ class DBAPI(DBConnection):
                 else:
                     desc = False
                 assert sqlbuilder.sqlIdentifier(s), "Strings in clauses are expected to be column identifiers.  I got: %r" % s
-                if select.sourceClass._SO_columnDict.has_key(s):
-                    s = select.sourceClass._SO_columnDict[s].dbName
+                if s in select.sourceClass.sqlmeta.columns:
+                    s = select.sourceClass.sqlmeta.columns[s].dbName
                 if desc:
                     return sqlbuilder.DESC(sqlbuilder.SQLConstant(s))
                 else:
@@ -326,17 +518,23 @@ class DBAPI(DBConnection):
     def _SO_dropJoinTable(self, join):
         self.query("DROP TABLE %s" % join.intermediateTable)
 
+    def _SO_createIndex(self, soClass, index):
+        self.query(self.createIndexSQL(soClass, index))
+
+    def createIndexSQL(self, soClass, index):
+        assert 0, 'Implement in subclasses'
+
     def createTable(self, soClass):
         self.query(self.createTableSQL(soClass))
 
     def createTableSQL(self, soClass):
-        return ('CREATE TABLE %s (\n%s\n)' % 
-                (soClass._table, self.createColumns(soClass)))
+        return ('CREATE TABLE %s (\n%s\n)' %
+                (soClass.sqlmeta.table, self.createColumns(soClass)))
 
     def createColumns(self, soClass):
         columnDefs = [self.createIDColumn(soClass)] \
                      + [self.createColumn(soClass, col)
-                        for col in soClass._SO_columns]
+                        for col in soClass.sqlmeta.columnList]
         return ",\n".join(["    %s" % c for c in columnDefs])
 
     def createColumn(self, soClass, col):
@@ -352,6 +550,13 @@ class DBAPI(DBConnection):
         # that.
         self.query("DELETE FROM %s" % tableName)
 
+    def createBinary(self, value):
+        """
+        Create a binary object wrapper for the given database.
+        """
+        # Default is Binary() function from the connection driver.
+        return self.module.Binary(value)
+
     # The _SO_* series of methods are sorts of "friend" methods
     # with SQLObject.  They grab values from the SQLObject instances
     # or classes freely, but keep the SQLObject class from accessing
@@ -359,42 +564,51 @@ class DBAPI(DBConnection):
     # in the SQLObject class.
 
     def _SO_update(self, so, values):
-        self.query("UPDATE %s SET %s WHERE %s = %s" %
-                   (so._table,
-                    ", ".join(["%s = %s" % (dbName, self.sqlrepr(value))
+        self.query("UPDATE %s SET %s WHERE %s = (%s)" %
+                   (so.sqlmeta.table,
+                    ", ".join(["%s = (%s)" % (dbName, self.sqlrepr(value))
                                for dbName, value in values]),
-                    so._idName,
+                    so.sqlmeta.idName,
                     self.sqlrepr(so.id)))
 
     def _SO_selectOne(self, so, columnNames):
-        return self.queryOne("SELECT %s FROM %s WHERE %s = %s" %
-                             (", ".join(columnNames),
-                              so._table,
-                              so._idName,
-                              self.sqlrepr(so.id)))
+        columns = ", ".join(columnNames)
+        if columns:
+            return self.queryOne(
+                "SELECT %s FROM %s WHERE %s = (%s)" %
+                (columns,
+                 so.sqlmeta.table,
+                 so.sqlmeta.idName,
+                 self.sqlrepr(so.id)))
+        else:
+            return self.queryOne(
+                "SELECT NULL FROM %s WHERE %s = (%s)" %
+                (so.sqlmeta.table,
+                 so.sqlmeta.idName,
+                 self.sqlrepr(so.id)))
 
     def _SO_selectOneAlt(self, cls, columnNames, column, value):
         return self.queryOne("SELECT %s FROM %s WHERE %s = %s" %
                              (", ".join(columnNames),
-                              cls._table,
+                              cls.sqlmeta.table,
                               column,
                               self.sqlrepr(value)))
 
     def _SO_delete(self, so):
-        self.query("DELETE FROM %s WHERE %s = %s" %
-                   (so._table,
-                    so._idName,
+        self.query("DELETE FROM %s WHERE %s = (%s)" %
+                   (so.sqlmeta.table,
+                    so.sqlmeta.idName,
                     self.sqlrepr(so.id)))
 
     def _SO_selectJoin(self, soClass, column, value):
-        return self.queryAll("SELECT %s FROM %s WHERE %s = %s" %
-                             (soClass._idName,
-                              soClass._table,
+        return self.queryAll("SELECT %s FROM %s WHERE %s = (%s)" %
+                             (soClass.sqlmeta.idName,
+                              soClass.sqlmeta.table,
                               column,
                               self.sqlrepr(value)))
 
     def _SO_intermediateJoin(self, table, getColumn, joinColumn, value):
-        return self.queryAll("SELECT %s FROM %s WHERE %s = %s" %
+        return self.queryAll("SELECT %s FROM %s WHERE %s = (%s)" %
                              (getColumn,
                               table,
                               joinColumn,
@@ -402,7 +616,7 @@ class DBAPI(DBConnection):
 
     def _SO_intermediateDelete(self, table, firstColumn, firstValue,
                                secondColumn, secondValue):
-        self.query("DELETE FROM %s WHERE %s = %s AND %s = %s" %
+        self.query("DELETE FROM %s WHERE %s = (%s) AND %s = (%s)" %
                    (table,
                     firstColumn,
                     self.sqlrepr(firstValue),
@@ -419,15 +633,64 @@ class DBAPI(DBConnection):
                     self.sqlrepr(secondValue)))
 
     def _SO_columnClause(self, soClass, kw):
-        return ' AND '.join(['%s = %s' %
-                         (soClass._SO_columnDict[key].dbName,
-                          self.sqlrepr(value))
-                         for key, value
-                         in kw.items()])
+        ops = {None: "IS"}
+        data = {}
+        if 'id' in kw:
+            data[soClass.sqlmeta.idName] = popKey(kw, 'id')
+        for key, col in soClass.sqlmeta.columns.items():
+            if key in kw:
+                data[col.dbName] = popKey(kw, key)
+            elif col.foreignName in kw:
+                obj = popKey(kw, col.foreignName)
+                if isinstance(obj, main.SQLObject):
+                    data[col.dbName] = obj.id
+                else:
+                    data[col.dbName] = obj
+        if kw:
+            # pick the first key from kw to use to raise the error,
+            raise TypeError, "got an unexpected keyword argument(s): %r" % kw.keys()
+
+        if not data:
+            return None
+        return ' AND '.join(
+            ['%s %s %s' %
+             (dbName, ops.get(value, "="), self.sqlrepr(value))
+             for dbName, value
+             in data.items()])
 
     def sqlrepr(self, v):
         return sqlrepr(v, self.dbName)
 
+    def __del__(self):
+        self.close()
+
+    def close(self):
+        if not hasattr(self, '_pool'):
+            # Probably there was an exception while creating this
+            # instance, so it is incomplete.
+            return
+        if not self._pool:
+            return
+        self._poolLock.acquire()
+        try:
+            conns = self._pool[:]
+            self._pool[:] = []
+            for conn in conns:
+                try:
+                    conn.close()
+                except self.module.Error:
+                    pass
+            del conn
+            del conns
+        finally:
+            self._poolLock.release()
+
+    def createEmptyDatabase(self):
+        """
+        Create an empty database.
+        """
+        raise NotImplementedError
+
 class Iteration(object):
 
     def __init__(self, dbconn, rawconn, select, keepConnection=False):
@@ -449,6 +712,8 @@ class Iteration(object):
         if result is None:
             self._cleanup()
             raise StopIteration
+        if result[0] is None:
+            return None
         if self.select.ops.get('lazyColumns', 0):
             obj = self.select.sourceClass.get(result[0], connection=self.dbconn)
             return obj
@@ -467,20 +732,21 @@ class Iteration(object):
 
     def __del__(self):
         self._cleanup()
-    
-
 
 class Transaction(object):
 
     def __init__(self, dbConnection):
+        # this is to skip __del__ in case of an exception in this __init__
+        self._obsolete = True
         self._dbConnection = dbConnection
         self._connection = dbConnection.getConnection()
         self._dbConnection._setAutoCommit(self._connection, 0)
         self.cache = CacheSet(cache=dbConnection.doCache)
+        self._deletedCache = {}
         self._obsolete = False
 
     def assertActive(self):
-        assert not self._obsolete, "This transaction has already gone through COMMIT/ROLLBACK; create another transaction"
+        assert not self._obsolete, "This transaction has already gone through ROLLBACK; begin another transaction"
 
     def query(self, s):
         self.assertActive()
@@ -506,9 +772,17 @@ class Transaction(object):
         # still iterating through the results.
         # @@: But would it be okay for psycopg, with threadsafety
         # level 2?
-        return iter(list(Iteration(self, self._connection,
+        return iter(list(select.IterationClass(self, self._connection,
                                    select, keepConnection=True)))
 
+    def _SO_delete(self, inst):
+        cls = inst.__class__.__name__
+        if not self._deletedCache.has_key(cls):
+            self._deletedCache[cls] = []
+        self._deletedCache[cls].append(inst.id)
+        meth = new.instancemethod(self._dbConnection._SO_delete.im_func, self, self.__class__)
+        return meth(inst)
+
     def commit(self):
         if self._obsolete:
             # @@: is it okay to get extraneous commits?
@@ -516,6 +790,13 @@ class Transaction(object):
         if self._dbConnection.debug:
             self._dbConnection.printDebug(self._connection, '', 'COMMIT')
         self._connection.commit()
+        subCaches = [(sub[0], sub[1].allIDs()) for sub in self.cache.allSubCachesByClassNames().items()]
+        subCaches.extend([(x[0], x[1]) for x in self._deletedCache.items()])
+        for cls, ids in subCaches:
+            for id in ids:
+                inst = self._dbConnection.cache.tryGetByName(id, cls)
+                if inst is not None:
+                    inst.expire()
 
     def rollback(self):
         if self._obsolete:
@@ -544,29 +825,95 @@ class Transaction(object):
         try:
             func = attr.im_func
         except AttributeError:
-            return attr
+            if isinstance(attr, ConnWrapper):
+                return ConnWrapper(attr._soClass, self)
+            else:
+                return attr
         else:
             meth = new.instancemethod(func, self, self.__class__)
             return meth
 
     def _makeObsolete(self):
         self._obsolete = True
+        if self._dbConnection.autoCommit:
+            self._dbConnection._setAutoCommit(self._connection, 1)
         self._dbConnection.releaseConnection(self._connection,
                                              explicit=True)
         self._connection = None
+        self._deletedCache = {}
 
     def begin(self):
         # @@: Should we do this, or should begin() be a no-op when we're
         # not already obsolete?
-        assert self._obsolete, "You cannot begin a new transaction session without committing or rolling back the transaction"
+        assert self._obsolete, "You cannot begin a new transaction session without rolling back this one"
         self._obsolete = False
         self._connection = self._dbConnection.getConnection()
+        self._dbConnection._setAutoCommit(self._connection, 0)
 
     def __del__(self):
         if self._obsolete:
             return
         self.rollback()
 
+class ConnectionHub(object):
+
+    """
+    This object serves as a hub for connections, so that you can pass
+    in a ConnectionHub to a SQLObject subclass as though it was a
+    connection, but actually bind a real database connection later.
+    You can also bind connections on a per-thread basis.
+
+    You must hang onto the original ConnectionHub instance, as you
+    cannot retrieve it again from the class or instance.
+
+    To use the hub, do something like::
+
+        hub = ConnectionHub()
+        class MyClass(SQLObject):
+            _connection = hub
+
+        hub.threadConnection = connectionFromURI('...')
+
+    """
+
+    def __init__(self):
+        self.threadingLocal = threading_local()
+
+    def __get__(self, obj, type=None):
+        # I'm a little surprised we have to do this, but apparently
+        # the object's private dictionary of attributes doesn't
+        # override this descriptor.
+        if obj and obj.__dict__.has_key('_connection'):
+            return obj.__dict__['_connection']
+        return self.getConnection()
+
+    def __set__(self, obj, value):
+        obj.__dict__['_connection'] = value
+
+    def getConnection(self):
+        try:
+            return self.threadingLocal.connection
+        except AttributeError:
+            try:
+                return self.processConnection
+            except AttributeError:
+                raise AttributeError(
+                    "No connection has been defined for this thread "
+                    "or process")
+
+    def _set_threadConnection(self, value):
+        self.threadingLocal.connection = value
+
+    def _get_threadConnection(self):
+        return self.threadingLocal.connection
+
+    def _del_threadConnection(self):
+        del self.threadingLocal.connection
+
+    threadConnection = property(_get_threadConnection,
+                                _set_threadConnection,
+                                _del_threadConnection)
+
 class ConnectionURIOpener(object):
 
     def __init__(self):
@@ -591,13 +938,18 @@ class ConnectionURIOpener(object):
             assert inst.name.find(':') == -1, "You cannot include ':' in your class names (%r)" % cls.name
             self.instanceNames[inst.name] = inst
 
-    def connectionForURI(self, uri):
+    def connectionForURI(self, uri, **args):
+        if args:
+            if '?' not in uri:
+                uri += '?'
+            uri += urllib.urlencode(args)
         if self.cachedURIs.has_key(uri):
             return self.cachedURIs[uri]
         if uri.find(':') != -1:
             scheme, rest = uri.split(':', 1)
-            assert self.schemeBuilders.has_key(scheme), \
-                   "No SQLObject driver exists for %s" % scheme
+            assert self.schemeBuilders.has_key(scheme), (
+                   "No SQLObject driver exists for %s (only %s)"
+                   % (scheme, ', '.join(self.schemeBuilders.keys())))
             conn = self.schemeBuilders[scheme]().connectionFromURI(uri)
         else:
             # We just have a name, not a URI
